<?php

namespace Drupal\redirect_bulk\Form;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Url;
use Drupal\redirect\Entity\Redirect;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form for adding bulk redirects.
 */
class RedirectBulkForm extends FormBase {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The language manager service.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Constructs a RedirectBulkForm object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $languageManager
   *   The language manager service.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, LanguageManagerInterface $languageManager) {
    $this->entityTypeManager = $entityTypeManager;
    $this->languageManager = $languageManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('language_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'redirect_bulk_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Initialize the number of redirects to add.
    $num_redirects = $form_state->get('num_redirects') ?? 1;

    // Wrapper for AJAX functionality.
    $form['redirects'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Redirects'),
      '#prefix' => '<div id="redirects-wrapper">',
      '#suffix' => '</div>',
    // Important to handle nested form values correctly.
      '#tree' => TRUE,
    ];

    // Add fields for each redirect.
    for ($i = 0; $i < $num_redirects; $i++) {
      $form['redirects'][$i] = [
        '#type' => 'container',
        '#attributes' => ['class' => ['redirect-entry']],
      ];

      // Source URL field.
      $form['redirects'][$i]['source'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Path'),
        '#required' => TRUE,
        '#default_value' => '',
        '#attributes' => ['class' => ['redirect-source']],
      ];

      // Destination URL field with autocomplete.
      $form['redirects'][$i]['destination'] = [
        '#type' => 'textfield',
        '#title' => $this->t('To'),
        '#required' => TRUE,
        '#default_value' => '',
        '#attributes' => ['class' => ['redirect-destination']],
        '#autocomplete_route_name' => 'redirect_bulk.node_autocomplete',
        '#autocomplete_route_parameters' => [],
      ];

      // Status code selection field.
      $form['redirects'][$i]['status_code'] = [
        '#type' => 'select',
        '#title' => $this->t('Redirect status'),
        '#options' => redirect_status_code_options(),
        '#required' => TRUE,
        '#default_value' => 301,
        '#attributes' => ['class' => ['redirect-status']],
      ];

      // Language selection field (multilingual support).
      if ($this->languageManager->isMultilingual()) {
        $form['redirects'][$i]['language'] = [
          '#type' => 'select',
          '#title' => $this->t('Language'),
          '#options' => [],
          '#default_value' => LanguageInterface::LANGCODE_NOT_SPECIFIED,
          '#attributes' => ['class' => ['redirect-language']],
        ];
        foreach ($this->languageManager->getLanguages(LanguageInterface::STATE_CONFIGURABLE) as $langcode => $language) {
          $form['redirects'][$i]['language']['#options'][$langcode] = $language->getName();
        }
        $form['redirects'][$i]['language']['#options'][LanguageInterface::LANGCODE_NOT_SPECIFIED] = $this->t('- All languages -');
      }

      // Remove button for each redirect entry.
      $form['redirects'][$i]['remove'] = [
        '#type' => 'submit',
        '#value' => $this->t('Remove'),
        '#name' => 'remove-' . $i,
        '#submit' => ['::removeRedirect'],
        '#ajax' => [
          'callback' => '::addmoreCallback',
          'wrapper' => 'redirects-wrapper',
          'event' => 'click',
          'progress' => [
            'type' => 'throbber',
            'message' => $this->t('Removing redirect...'),
          ],
        ],
        '#attributes' => ['class' => ['redirect-remove']],
        '#limit_validation_errors' => [],
      ];
    }

    // Add another redirect button.
    $form['add_redirect'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add redirect'),
      '#submit' => ['::addOne'],
      '#ajax' => [
        'callback' => '::addmoreCallback',
        'wrapper' => 'redirects-wrapper',
        'event' => 'click',
        'progress' => [
          'type' => 'throbber',
          'message' => $this->t('Adding more redirects...'),
        ],
      ],
      '#attributes' => ['class' => ['add-redirect']],
      '#button_type' => 'primary',
    ];

    // Submit button for the form.
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
      '#attributes' => ['class' => ['save-redirects']],
      '#button_type' => 'primary',
    ];

    // Add custom CSS for form layout.
    $form['#attached']['library'][] = 'redirect_bulk/redirect_bulk';

    return $form;
  }

  /**
   * AJAX callback to add or remove redirects.
   */
  public function addmoreCallback(array &$form, FormStateInterface $form_state) {
    return $form['redirects'];
  }

  /**
   * Submit handler for the "add one more redirect" button.
   */
  public function addOne(array &$form, FormStateInterface $form_state) {
    $num_redirects = $form_state->get('num_redirects') ?? 1;
    $form_state->set('num_redirects', $num_redirects + 1);
    $form_state->setRebuild();
  }

  /**
   * Submit handler for removing a redirect entry.
   */
  public function removeRedirect(array &$form, FormStateInterface $form_state) {
    // Get the index of the redirect to remove.
    $triggering_element = $form_state->getTriggeringElement();
    $index = substr($triggering_element['#name'], strlen('remove-'));

    // Remove the redirect entry from the form state.
    $redirects = $form_state->getValue(['redirects']);
    unset($redirects[$index]);
    $form_state->setValue(['redirects'], array_values($redirects));

    // Decrement the redirect counter.
    $num_redirects = $form_state->get('num_redirects') ?? 1;
    if ($num_redirects > 0) {
      $form_state->set('num_redirects', $num_redirects - 1);
    }

    // Rebuild the form.
    $form_state->setRebuild();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);

    $source_paths = [];
    $redirects = $form_state->getValue(['redirects']);

    foreach ($redirects as $key => $redirect) {
      $source = trim($redirect['source']);
      $destination = trim($redirect['destination']);

      // Validate required fields.
      if (empty($source)) {
        $form_state->setErrorByName("redirects][$key][source", $this->t('Source Path is required.'));
        continue;
      }

      if (empty($destination)) {
        $form_state->setErrorByName("redirects][$key][destination", $this->t('The url to redirect from is required.'));
        continue;
      }

      // Check for duplicate source URLs within the form.
      $parsed_url = UrlHelper::parse($source);
      $path = $parsed_url['path'] ?? NULL;
      $query = $parsed_url['query'] ?? NULL;
      $source_with_query = $path . '?' . UrlHelper::buildQuery($query);

      if (in_array($source_with_query, $source_paths)) {
        $form_state->setErrorByName("redirects][$key][source", $this->t('The source Path %source is duplicated in the form.', ['%source' => $source_with_query]));
        continue;
      }
      $source_paths[] = $source_with_query;

      // Additional validations.
      if ($source == '<front>') {
        $form_state->setErrorByName("redirects][$key][source", $this->t('It is not allowed to create a redirect from the front page.'));
        continue;
      }

      if (strpos($source ?? '', '#') !== FALSE) {
        $form_state->setErrorByName("redirects][$key][source", $this->t('Anchor fragments are not allowed in the source Path.'));
        continue;
      }

      if (strpos($source ?? '', '?') === 0) {
        $form_state->setErrorByName("redirects][$key][source", $this->t('This value should not be null.'));
        continue;
      }

      if (strpos($source ?? '', '/') === 0) {
        $form_state->setErrorByName("redirects][$key][source", $this->t('The source Path should not start with a forward slash (/).'));
        continue;
      }

      if (strpos($destination ?? '', '/') === 0) {
        $form_state->setErrorByName("redirects][$key][destination", $this->t('The url to redirect from should not start with a forward slash (/).'));
        continue;
      }

      if (!$this->isValidPath($source)) {
        $form_state->setErrorByName("redirects][$key][source", $this->t('The source path %source is invalid.', ['%source' => $source]));
      }

      try {
        // Validate source and destination URLs.
        $source_url = Url::fromUri('internal:/' . ltrim($source));
        $redirect_url = Url::fromUri('internal:/' . ltrim($destination));

        // Check if redirecting to itself.
        if ($source_url->toString() == $redirect_url->toString()) {
          $form_state->setErrorByName('bulk_redirects', $this->t('You are attempting to redirect the page to itself. This will result in an infinite loop.'));
          return;
        }
      }
      catch (\InvalidArgumentException $e) {
        // Handle exception if URLs are invalid.
      }

      // Determine language for hash generation.
      $language = LanguageInterface::LANGCODE_NOT_SPECIFIED;
      if ($this->languageManager->isMultilingual() && isset($redirect['language']) && $redirect['language'] !== LanguageInterface::LANGCODE_NOT_SPECIFIED) {
        $language = $redirect['language'];
      }
      // Generate hash for the source path, query parameters, and language.
      $hash = Redirect::generateHash($path, $query, $language);

      // Validate against existing redirects in the database using the hash.
      $existing_redirects = $this->entityTypeManager
        ->getStorage('redirect')
        ->loadByProperties(['hash' => $hash]);

      if (!empty($existing_redirects)) {
        $existing_redirect = reset($existing_redirects);
        $form_state->setErrorByName("redirects][$key][source", $this->t('The source path %source is already being redirected. Edit the existing redirect <a href="@edit-url">here</a>.', [
          '%source' => $source_with_query,
          '@edit-url' => $existing_redirect->toUrl('edit-form')->toString(),
        ]));
        continue;
      }
    }
  }

  /**
   * Checks if the given path is valid.
   *
   * @param string $path
   *   The path to validate.
   *
   * @return bool
   *   TRUE if the path is valid, FALSE otherwise.
   */
  protected function isValidPath($path) {
    try {
      if (strpos($path, 'http://') === 0 || strpos($path, 'https://') === 0 || strpos($path, '/') === 0) {
        Url::fromUri('internal:' . $path, ['absolute' => TRUE]);
      }
      else {
        Url::fromUserInput('/' . $path);
      }
      return TRUE;

    }
    catch (\Exception $e) {
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $redirects = $form_state->getValue(['redirects']);
    foreach ($redirects as $redirect) {
      $source = trim($redirect['source']);
      $destination = trim($redirect['destination']);
      $status_code = (int) $redirect['status_code'];
      $language = $redirect['language'] ?? LanguageInterface::LANGCODE_NOT_SPECIFIED;

      $source_query = UrlHelper::parse($source);

      if (UrlHelper::isExternal($destination)) {
        $destination_uri = $destination;
      }
      elseif (strpos($destination, 'node') !== FALSE) {
        // Check if "node" is part of the destination path.
        $destination_uri = 'entity:' . ltrim($destination, '/');
      }
      else {
        $destination_uri = 'internal:/' . ltrim($destination);
      }

      // Ensure the query array is set correctly.
      $query = $source_query['query'] ?? [];

      // Save each redirect as a Redirect entity.
      Redirect::create([
        'redirect_source' => [
          'path' => $source_query['path'],
          'query' => $query,
        ],
        'redirect_redirect' => [
          'uri' => $destination_uri,
          'title' => '',
        ],
        'status_code' => $status_code,
        'language' => $language,
      ])->save();
    }

    // Clear form values after submission.
    $form_state->setValues([]);
    $form_state->set('num_redirects', 1);

    // Display a success message.
    $this->messenger()->addStatus($this->t('Redirects have been added.'));
    // Redirect to appropriate route after saving.
    $form_state->setRedirect('redirect.list');
  }

}
